﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.EvolutionaryTracking {
  [Item("SymbolicExpressionTreeGenealogyGraph", "A genealogy graph for symbolic expression trees")]
  [StorableClass]
  public class SymbolicExpressionTreeGenealogyGraph : GenealogyGraph<SymbolicExpressionGenealogyGraphNode>, ISymbolicExpressionTreeGenealogyGraph {
    [Storable]
    private readonly Dictionary<ISymbolicExpressionTree, List<SymbolicExpressionGenealogyGraphNode>> nodeMap =
      new Dictionary<ISymbolicExpressionTree, List<SymbolicExpressionGenealogyGraphNode>>();

    public SymbolicExpressionTreeGenealogyGraph() { }

    [StorableConstructor]
    public SymbolicExpressionTreeGenealogyGraph(bool serializing) : base(serializing) { }

    public override IDeepCloneable Clone(Cloner cloner) { return new SymbolicExpressionTreeGenealogyGraph(this, cloner); }

    protected SymbolicExpressionTreeGenealogyGraph(SymbolicExpressionTreeGenealogyGraph original, Cloner cloner) : base(original, cloner) { }

    public override void AddNode(SymbolicExpressionGenealogyGraphNode node) {
      if (!nodeMap.ContainsKey(node.SymbolicExpressionTree))
        nodeMap[node.SymbolicExpressionTree] = new List<SymbolicExpressionGenealogyGraphNode> { node };
      else
        nodeMap[node.SymbolicExpressionTree].Add(node);
      base.AddNode(node);
    }

    public List<SymbolicExpressionGenealogyGraphNode> GetGraphNodes(ISymbolicExpressionTree tree) {
      return nodeMap.ContainsKey(tree) ? nodeMap[tree] : null;
    }

    public bool HasNode(SymbolicExpressionTree tree) {
      return Nodes.Any(x => x.SymbolicExpressionTree == tree);
    }

    public int Compare(GenealogyGraphNode a, GenealogyGraphNode b) {
      return Compare(a as SymbolicExpressionGenealogyGraphNode, b as SymbolicExpressionGenealogyGraphNode);
    }

    // TODO: make sure a and b are valid and the info exists for both
    public int Compare(SymbolicExpressionGenealogyGraphNode a, SymbolicExpressionGenealogyGraphNode b) {
      if (a.Quality.Equals(b.Quality)) {
        if (a.IsElite) return +1;
        if (b.IsElite) return -1;
      }
      return a.Quality.CompareTo(b.Quality);
    }

    #region Fragment tracing
    public IEnumerable<ISymbolicExpressionTree> TraceFragment(ISymbolicExpressionTreeNode fragment, SimilarityLevel similarity = SimilarityLevel.Exact) {
      return Nodes.Where(x => x.SymbolicExpressionTree.ContainsFragment(fragment, similarity)).Select(x => x.SymbolicExpressionTree).ToList();
    }
    #endregion
  }

  public class SymbolicExpressionGenealogyGraphNode : GenealogyGraphNode {
    public ISymbolicExpressionTree SymbolicExpressionTree { get; set; }
    public double Quality { get; set; }
    public bool IsElite { get; set; }
    public List<double> Ranks { get; set; }

    public SymbolicExpressionGenealogyGraphNode() {
      Ranks = new List<double>();
    }

    public SymbolicExpressionGenealogyGraphNode(object content) {
      Content = content;
      Ranks = new List<double>();
    }
  }
}
